"""
This probably isn't optmized/the minimum amount of code to do this
Dealing with sockets/EOFs/random drops is a pain, so this stays on the safe side
of validating to make sure sockets arent left dangling or data isn't
saved, because of a ConnectionRefusedError/BrokenPipe/general OSErrors

logzero logs all the exceptions, incase theyre not what I expect
most of the times, mpv will be open for more than 10 minutes, so
the WRITE_PERIOD periodically writes will at least capture what was
being listened to, even if *somehow* (has only happened
when my computer suffers a random crash/shut-down)
I lose data on what happened
when mpv EOFd/disconnected due to a BrokenPipe.
BrokenPipes are captured in the event_eof function
"""

import os
import json
import atexit
from pathlib import Path
from typing import List, Optional, Dict, Any
from time import sleep, time

from python_mpv_jsonipc import MPV  # type: ignore[import]
from logzero import logger, logfile  # type: ignore[import]

SCAN_TIME: int = 10


KNOWN_EVENTS = set(
    [
        "socket-added",  # custom event, for when the socket was added
        "mpv-quit",
        "playlist-count",
        "is-paused",
        "eof",
        "seek",
        "paused",
        "resumed",
        "metadata",
        "duration",
        "playlist-pos",
        "media-title",
        "path",
        "working-directory",
        "final-write",  # custom event, for when the dead/dangling socket was removed, and file was written
    ]
)

# every 10 minutes
WRITE_PERIOD = 600


def new_event(event_name: str, event_data: Any = None) -> Dict[str, Any]:
    """
    helper to create an event. validates the event name
    """
    if event_name not in KNOWN_EVENTS:
        logger.warning(f"Unknown event: {event_name}")
    return {event_name: event_data}


# disabled for now
def clean_playlist(mpv_playlist_response: List[Dict]) -> List[str]:
    """
    simplifies the playlist response from mpv
    from:
        {'filename': '01 Donuts (Outro).mp3'}, {'filename': '02 Workinonit.mp3', 'current': True, 'playing': True},...
    to:
        ['01 Donuts (Outro).mp3', ...]
    """
    filenames: List[str] = []
    for pinfo in mpv_playlist_response:
        if "filename" not in pinfo:
            logger.warning(f"No filename in playlist info!: {pinfo}")
        else:
            filenames.append(pinfo["filename"])
    # truncate to 50 filenames
    return filenames[:50]


class SocketData:
    """
    Stores Metadata for a socket with timestamps in memory
    Writes out to the JSON file when the MPV process ends

    Want to save:
        # at mpv launch:
            # playlist
            # playlist-count
            # working-directory
            # is-paused
        # event-based
            # whenever a socket is played/paused
            # whenever a file changes (eof-reached) save metadata about whats being played
                # metadata
                # media-title
                # path
                # playlist-pos
                # duration
    """

    def __init__(self, socket: MPV, socket_loc: str, data_dir: str):
        self.socket = socket
        self.socket_loc = socket_loc
        self.data_dir = data_dir
        self.socket_time = socket_loc.split("/")[-1]
        self.events: Dict[float, Dict] = {}
        # write every 10 minutes, even if mpv doesnt exit
        self.write_at = time() + WRITE_PERIOD
        # keep track of playlist/playlist-count, so we can use eof to determine
        # whether we should read next metadata
        self.playlist_count = self.socket.playlist_count
        playlist_pos = self.socket.playlist_pos
        # incremented at the top of in store_file_metadata
        # technically this is zero-indexed, but have to deal with off-by-one errors
        # because I cant read the playlist position of a dead socket, counting manually
        if playlist_pos is None:
            logger.warning(
                "Couldn't get playlist position in SocketData initialization, defaulting to 0"
            )
            self.playlist_index = 0
        else:
            self.playlist_index = playlist_pos
        self.store_initial_metadata()
        self.store_file_metadata()

    @property
    def event_count(self):
        return len(self.events)

    _repr_attrs = ("socket", "socket_loc", "event_count")

    def __repr__(self) -> str:
        return "{}({})".format(
            self.__class__.__name__,
            ", ".join(
                [
                    "=".join([a, str(getattr(self, a))])
                    for a in self.__class__._repr_attrs
                ]
            ),
        )

    __str__ = __repr__

    def write(self):
        with open(
            os.path.join(self.data_dir, f"{self.socket_time}.json"), "w"
        ) as event_f:
            json.dump(self.events, event_f)

    def nevent(self, event_name: str, event_data: Optional[Any] = None) -> None:
        """add an event"""
        ct = time()
        logger.debug(f"{self.socket_time}|{ct}|{event_name}|{event_data}")
        self.events[ct] = new_event(event_name, event_data)

    def store_initial_metadata(self):
        self.nevent("socket-added", time())
        self.nevent("working-directory", self.socket.working_directory)
        self.nevent("playlist-count", self.socket.playlist_count)
        self.nevent("is-paused", self.socket.pause)
        # self.nevent("playlist", clean_playlist(self.socket.playlist))

    def poll_for_property(
        self, attr: str, event_name: str, tries: int = 20, create_event: bool = True
    ) -> Any:
        """
        Some properties arent set when the file starts?
        sleeps for 0.1 of a second between tries (20 * 0.1 = 2 seconds)

        if create_event, once it has a non-None value, it sets the value
        with self.nevent
        """
        for _ in range(tries):
            value = getattr(self.socket, attr)
            if value is not None:
                if create_event:
                    self.nevent(event_name, value)
                return value
            sleep(0.1)
        else:
            logger.warning(f"{self.socket_loc} Couldn't poll for {event_name}")

    def store_file_metadata(self):
        """
        Called when EOF is reached (so, another file starts)
        """

        self.playlist_index += 1
        # (is zero indexed, but incrementing before making any requests out to the socket)
        if self.playlist_index - 1 >= self.playlist_count:
            logger.debug("Reached end of playlist, not reading in next file info...")
            return

        # poll for these incase theyre not set for some reason, because
        # the file was just loaded by mpv
        actual_playlist_pos = self.poll_for_property("playlist_pos", "playlist-pos")
        # make sure internal, manually counted playlist index is accurate
        if actual_playlist_pos is not None:
            if actual_playlist_pos + 1 != self.playlist_index:
                self.playlist_index = actual_playlist_pos + 1
        self.poll_for_property("path", "path")
        self.poll_for_property("media_title", "media-title")

        # weird, metadata and duration arent received at the beginning of the file?
        # poll for duration and metadata
        # maybe these have to be parsed and there done a bit after the file is read
        self.poll_for_property("metadata", "metadata")
        self.poll_for_property("duration", "duration")

    def event_resumed(self) -> None:
        """
        Called when the media is resumed, also save % in file
        """
        self.nevent("resumed", {"percent-pos": self.socket.percent_pos})

    def event_paused(self) -> None:
        """
        Called when the media is paused, also save % in file
        """
        self.nevent("paused", {"percent-pos": self.socket.percent_pos})

    def event_eof(self):
        """
        Called when the 'eof' event happens. Doesn't necessarily mean mpv exits
        Could be going to the next song in the current playlist
        (This doesnt happen the first time a song is loaded)

        Though, this is also called when mpv exits, so we should wrap the
        possible socket errors
        """
        self.nevent("eof")
        try:
            self.store_file_metadata()  # store info about new file thats playing
        # possible errors thrown: BrokenPipeError, OSError
        except Exception as e:
            logger.warning(f"Ignoring error: {e}")
            if not isinstance(e, ConnectionRefusedError):
                logger.exception(e)

    def event_seeking(self):
        """
        Called when the user seeks in the file. Could possibly be called when a file is loaded as well
        """
        if self.socket.percent_pos < 2:
            # logger.debug("ignoring seek because we just EOFd?")
            pass
        else:
            # save what % we seeked to
            self.nevent("seek", {"percent-pos": self.socket.percent_pos})


class LoopHandler:
    """
    Handles keeping track of currently live mpv instances, attaching handlers,
    the main loop.
    """

    def __init__(self, socket_dir: str, data_dir: str, autostart: bool = True):
        self.data_dir: str = data_dir
        self.socket_dir: str = socket_dir
        self._socket_dir_path: Path = Path(socket_dir).expanduser().absolute()
        self.sockets: Dict[str, MPV] = {}
        self.socket_data: Dict[str, SocketData] = {}
        if autostart:
            self.run_loop()

    def scan_sockets(self):
        """
        Look for any new sockets at socket_dir, remove any dead ones
        """
        socket_loc: str = None
        try:
            # iterate through all files
            for socket_name in os.listdir(self.socket_dir):
                socket_loc: str = os.path.join(self.socket_dir, socket_name)
                if socket_loc not in self.sockets:
                    # each of these runs in a separate thread, so the while loop below doesnt block event data
                    # ConnectionRefusedError thrown here
                    new_sock = MPV(
                        start_mpv=False,
                        ipc_socket=socket_loc,
                        quit_callback=lambda: self.remove_socket(socket_loc),
                    )
                    self.sockets[socket_loc] = new_sock
                    # if the socket gets disconnected for some reason, and we're recreating MPV, *never* overwrite data
                    if socket_loc in self.socket_data:
                        self.socket_data[socket_loc].socket = new_sock
                    else:
                        self.socket_data[socket_loc] = SocketData(
                            new_sock, socket_loc, self.data_dir
                        )
                    self.attach_observers(socket_loc, new_sock)
                    self.debug_internals()
                else:  # if this socket is already connected, just try to get the path from the scoket
                    # may have been a TimeoutError: No response from MPV.
                    # which resulted in the socket remaning in self.sockets, even if its eof'd and exited
                    self.sockets[socket_loc].path
            # iterate through sockets, if file doesnt exist for some reason
            # this is probably unnecessary
            for s_loc, sock_obj in self.sockets.items():
                # update higher scope to allow usage in except block
                socket_loc = s_loc
                # try to access path to possibly cause ConnectionRefusedError,
                # removing a dead socket
                sock_obj.path
        except (ConnectionRefusedError, BrokenPipeError):
            logger.debug(
                f"Connected refused for socket at {socket_loc}, removing dead/dangling socket file..."
            )
            # make sure its actually removed from active sockets
            # gets removed from socket_data after 10 seconds
            if socket_loc in self.sockets:
                self.remove_socket(socket_loc)
            # rm -f
            try:
                os.remove(socket_loc)
            except FileNotFoundError:
                pass

    def attach_observers(self, socket_loc: str, sock: MPV) -> None:
        """
        Watch for user pausing, eof-file (file ending)
        """
        # sanity-check
        if hasattr(sock, "_watching_instance"):
            logger.warning("Tried to attach observers twice!")
            return
        setattr(sock, "_watching_instance", None)

        socket_data: SocketData = self.socket_data[socket_loc]

        # keep track of when last EOF was. EOF also happens when
        # a file is loaded, and seeking happens when you load a file,
        # (since its sort of seeking to the beginning)
        # doesnt seem to be deterministic/easy to filter seeks out
        # by EOFs, and might match actual seeking. so, will have
        # to do larger analysis on the dumped data to figure out
        # if EOF next to seek, remove the seek

        @sock.property_observer("pause")
        def on_pause(_name, value):
            if value:  # item is now paused
                socket_data.event_paused()
            else:
                socket_data.event_resumed()

        @sock.property_observer("eof-reached")
        def on_eof(_name, value):
            # value == False means that eof has not been reached
            if isinstance(value, bool) and not value:
                return
            if value is not None:
                logger.warning(
                    "Seems that this is supposed to be None; just to signify event? not sure why it isnt"
                )
            socket_data.event_eof()

        @sock.property_observer("seeking")
        def on_seek(_name, value):
            if isinstance(value, bool) and value:
                socket_data.event_seeking()

    def remove_socket(self, socket_loc: str) -> None:
        if socket_loc in self.sockets:
            logger.debug(f"Removing socket {socket_loc}")
            del self.sockets[socket_loc]
            # write quit event
            self.socket_data[socket_loc].nevent("mpv-quit", time())
        else:
            logger.warning(
                "called remove socket, but socket_loc doesnt exist in self.sockets"
            )
        # (doesnt remove the file here, but should find it on the next scan_sockets call and remove it then)

    def debug_internals(self):
        logger.debug("sockets {}".format(self.sockets))
        logger.debug("socket_data {}".format(self.socket_data))

    def periodic_write(self):
        now = time()
        for socket_loc, socket_data in self.socket_data.items():
            if now > socket_data.write_at:
                logger.debug(f"{socket_data.socket_time}|running periodic write")
                socket_data.write()
                socket_data.write_at = now + WRITE_PERIOD
                self.debug_internals()

    def write_data(self, force: bool = False) -> None:
        """
        Write out any completed SocketData to disk
        """
        # if the /tmp/mpvsockets/ file is no longer in self.sockets
        # but we have socketdata for it from when it was alive, in
        # self.socket_data, write that out to data_dir
        #
        # runs in the main thread, errors crash main thread
        for socket_loc in list(self.socket_data):
            if socket_loc not in self.sockets:
                logger.info(f"{socket_loc}: writing to file...")
                self.socket_data[socket_loc].nevent("final-write", time())
                self.socket_data[socket_loc].write()
                del self.socket_data[socket_loc]
                self.debug_internals()
        if force:
            # don't write additional devents to the file, just write data
            # for every socket regardless of state. this is used if the program
            # is crashing/etc.
            logger.warning("forcing write to files...")
            self.debug_internals()
            for socket_data in self.socket_data.values():
                socket_data.write()

    def run_loop(self):
        logger.debug("Starting mpv-history-daemon loop...")
        while True:
            self.scan_sockets()
            self.periodic_write()
            self.write_data()
            sleep(SCAN_TIME)


def run(socket_dir: str, data_dir: str, log_file: str) -> None:
    # if the daemon launched before any mpv instances
    if not os.path.exists(socket_dir):
        os.makedirs(socket_dir)
    assert os.path.isdir(socket_dir)
    os.makedirs(data_dir, exist_ok=True)
    assert os.path.isdir(data_dir)
    logfile(log_file, maxBytes=1e7, backupCount=1)
    l = LoopHandler(socket_dir, data_dir, autostart=False)
    # incase user keyboardinterrupt's or this crashes completely
    # for some reason, write data out to files in-case it hasn't
    # been done recently
    atexit.register(lambda: l.write_data(force=True))
    l.run_loop()



import os
import sys
import datetime
from pathlib import Path
from typing import Any
from tempfile import gettempdir

import click
import simplejson
from logzero import setup_logger  # type: ignore[import]

from .daemon import run
from .events import history, all_history
from .events import logger as event_logger


@click.group()
def cli():
    """
    Connects to mpv socket files and saves a history of events
    """
    pass


@cli.command()
@click.argument("SOCKET_DIR")
@click.argument("DATA_DIR")
@click.option(
    "--log-file",
    type=click.Path(),
    default=os.path.join(gettempdir(), "mpv-history-daemon.log"),
    help="location of logfile",
)
def daemon(socket_dir: str, data_dir: str, log_file: str) -> None:
    """
    Socket dir is the directory with mpv sockets (/tmp/mpvsockets, probably)
    Data dir is the directory to store the history JSON files
    """
    run(socket_dir, data_dir, log_file)


def default_encoder(o: Any) -> Any:
    if isinstance(o, datetime.datetime):
        return int(o.timestamp())
    raise TypeError(f"{o} of type {type(o)} is not serializable")


@cli.command()
@click.argument("DATA_DIR")
@click.option(
    "--all-events",
    is_flag=True,
    default=False,
    help="return all events, even ones which by context you probably didn't listen to",
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="Increase log verbosity/print warnings while parsing JSON files",
)
def parse(data_dir: str, all_events: bool, debug: bool) -> None:
    """
    Takes the data directory and parses events into Media
    """
    global event_logger
    if debug:
        event_logger = setup_logger(__name__, level="DEBUG")
    events_func: Any = all_history if all_events else history
    ddir: Path = Path(data_dir)
    if not ddir.exists():
        click.echo(f"{data_dir} does not exist", err=True)
        sys.exit(1)
    if not ddir.is_dir():
        click.echo(f"{data_dir} is not a directory", err=True)
        sys.exit(1)
    click.echo(
        simplejson.dumps(
            list(events_func(list(ddir.iterdir()))),
            default=default_encoder,
            namedtuple_as_object=True,
        )
    )


if __name__ == "__main__":
    cli(prog_name="mpv-history-daemon")



"""
Reads the JSON event files and constructs media files
"""

import os
import re
import json
from itertools import chain
from datetime import datetime, timezone
from pathlib import Path
from typing import (
    Iterator,
    Sequence,
    List,
    NamedTuple,
    Set,
    Any,
    Dict,
    Tuple,
    Optional,
    Union,
    Callable,
)

from logzero import setup_logger  # type: ignore[import]

from .daemon import SCAN_TIME

# TODO: better logger setup?
loglevel: str = os.environ.get("MPV_HISTORY_EVENTS_LOGLEVEL", "info").upper()
logger = setup_logger("mpv_history_events", level=loglevel)

EventType = str
EventData = Any


def parse_datetime_sec(d: Union[str, float, int]) -> datetime:
    return datetime.fromtimestamp(int(d), tz=timezone.utc)


# changed to 'since_started' rather than using a timedelta
# so this is cachew compliant
# see https://github.com/karlicoss/cachew/issues/28
class Action(NamedTuple):
    # this event happened this many seconds after this media was started
    since_started: float
    action: str
    percentage: float


class Media(NamedTuple):
    path: str  # local or URL path
    is_stream: bool  # if streaming from a URL
    start_time: datetime  # when the media started playing
    end_time: datetime  # when the media was closed/finished
    pause_duration: float  # how long the media was paused for (typically 0)
    media_duration: Optional[float]  # length of the media
    # title of the media (if URL, could be <title>...</title> from ytdl
    media_title: Optional[str]
    # additional metadata on what % I was through the media while pausing/playing/seeking
    actions: List[Action]
    metadata: Dict[str, Any]  # metadata from the file, if it exists

    @property
    def score(self) -> float:
        """Describes how much data this piece of media has, to resolve conflicts"""
        sc = 0
        if self.media_title is not None:
            sc = sc + 1
        if self.media_duration is not None:
            sc = sc + 1
        if self.pause_duration > 1.0:
            sc = sc + 1
        sc = sc + int(len(self.metadata) / 4)
        sc = sc + int(len(self.actions) / 8)
        return float(sc)

    @property
    def listen_time(self) -> float:
        return (self.end_time - self.start_time).total_seconds() - self.pause_duration


Results = Iterator[Media]


def all_history(input_files: Sequence[Path]) -> Results:
    yield from chain(*map(_read_event_stream, input_files))


# use some of the context of what this piece of media
# is to figure out if I actually watched/listened to it.
# I may have skipped a song if it only has a couple
# seconds between when it started/ended
def _actually_listened_to(m: Media) -> bool:
    listen_time: float = m.listen_time
    # if this is mpv streaming something from /dev/
    # (like my camera), ignore
    if not m.is_stream and m.path.startswith("/dev/"):
        return False
    if m.media_duration is not None and m.media_duration != 0:
        percentage_listened_to = listen_time / m.media_duration
        # if under 10 minutes (probably a song?), if I listened to at least 75%
        if m.media_duration < 600:
            return percentage_listened_to > 0.75
    # otherwise, just check if return if I listened to at least a minute
    return listen_time > 60


# filter out items I probably didn't listen to
def history(
    input_files: Sequence[Path],
    filter_function: Callable[[Media], bool] = _actually_listened_to,
) -> Results:
    """
    can supply a function which accepts a 'Media' object as
    the first argument as the filter function
    """
    yield from filter(filter_function, all_history(input_files))


def _read_event_stream(p: Path) -> Results:
    # if theres a conflict, keep a 'score' by adding non-null fields on an item,
    # and return the one that has the most
    #
    # sometimes youtube-dl will show up twice ...?
    # use 'path' as a primary key to remove possible
    # duplicate event data
    items: Dict[str, Media] = {}
    for d in _reconstruct_event_stream(p):
        # required keys
        if not REQUIRED_KEYS.issubset(set(d)):
            # logger.debug("Doesnt have required keys, ignoring...")
            continue
        if d["end_time"] < d["start_time"]:
            logger.warning(f"End time is less than start time! {d}")
        fdur: Optional[float] = None
        if "duration" in d:
            fdur = float(d["duration"])
        start_time = parse_datetime_sec(int(d["start_time"]))
        m = Media(
            path=d["path"],
            is_stream=d["is_stream"],
            start_time=start_time,
            end_time=parse_datetime_sec(int(d["end_time"])),
            pause_duration=float(d["pause_duration"]),
            media_duration=fdur,
            media_title=d.get("media_title"),
            actions=[
                Action(
                    since_started=(
                        parse_datetime_sec(timestamp) - start_time
                    ).total_seconds(),
                    action=data[0],
                    percentage=data[1],
                )
                for timestamp, data in d["actions"].items()
            ],
            metadata=d.get("metadata", {}),
        )
        key = m.path
        if key not in items:
            items[key] = m
        else:
            # use item with better score
            if m.score > items[key].score:
                logger.debug(f"replacing {items[key]} with {m}")
                items[key] = m
    yield from list(items.values())


REQUIRED_KEYS = set(["playlist_pos", "start_time", "path"])

IGNORED_EVENTS: Set[EventType] = set(
    [
        "playlist",
        "playlist-count",
    ]
)


def _destructure_event(d: Dict) -> Tuple[EventType, EventData]:
    di = d.items()
    assert len(di) == 1, "Event not in expected format!"
    return list(di)[0]


URL_REGEX = re.compile(
    r"^(?:http|ftp)s?://"  # http:// or https://
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain...
    r"localhost|"  # localhost...
    r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
    r"(?::\d+)?"  # optional port
    r"(?:/?|[/?]\S+)$",
    re.IGNORECASE,
)


# https://stackoverflow.com/a/7160778/9348376
def _is_urlish(url: str) -> bool:
    return re.match(URL_REGEX, url) is not None


def _reconstruct_event_stream(p: Path) -> Iterator[Dict[str, Any]]:
    """
    Takes about a dozen events receieved chronologically from the MPV
    socket, and reconstructs what I was doing while it was playing.
    """
    events = json.loads(p.read_text())
    # mpv socket names are created like:
    #
    # declare -a mpv_options
    # mpv_options=(--input-ipc-server="${socket_dir}/$(date +%s%N)")
    # exec "$mpv_path" "${mpv_options[@]}"
    #
    # get when mpv launched from the filename
    start_time = None
    try:
        start_time = int(int(p.stem) / 1e9)
    except ValueError as ve:
        logger.warning(str(ve))

    # dictionary for storing data while we parse though events
    media_data: Dict[str, Any] = {}

    # 'globals', set at the beginning
    working_dir = os.environ["HOME"]
    is_first_item = True  # helps control how to handle duration
    # playlist_count = None
    most_recent_time: float = 0.0

    # used to help determine state
    is_playing = True  # assume playing at beginning
    pause_duration = 0.0  # pause duration for this entry
    pause_start_time: Optional[float] = None  # if the entry is paused, when it started
    actions: Dict[float, Tuple[str, float]] = {}

    # sort by timestamp, incase
    for dt_s in sorted(events):
        dt_float = float(dt_s)
        most_recent_time = dt_float
        event_name, event_data = _destructure_event(events[dt_s])
        if event_name in IGNORED_EVENTS:
            continue
        elif event_name == "playlist-pos":
            # reliable event to use to set start time of an item
            # the first item might have been off for 5 or so seconds
            # because of the socket_scan, so we cant use playlist-pos's
            # timestamp as the start of the mpv instance.
            # instead, we use the timestamp from the /tmp/mpvsocket/ filename
            #
            # but, if this is not the first item in the event stream,
            # use playlist-pos's timestamp as when a file starts
            if (
                "playlist_pos" in media_data
                and media_data["playlist_pos"] == event_data
            ):
                logger.debug(
                    f"Got same playlist position {event_data} twice. Current data: {media_data}"
                )
                continue
            media_data["playlist_pos"] = event_data
            if is_first_item:
                # if this is the first item, set the start time to when mpv launched
                media_data["start_time"] = start_time
                is_first_item = False  # stays false the entire function call
            else:
                media_data["start_time"] = dt_float
        elif event_name == "socket-added":
            if start_time is None:
                start_time = int(float(event_data))
        elif event_name == "working-directory":
            # shouldnt be added to media_data, affects path, but is the
            # same across the entire run of mpv
            working_dir = event_data
        elif event_name == "is-paused":
            # if this was paused when we connected to the socket,
            # assume its been paused since close to it was launched
            if event_data is True:
                is_playing = False
                pause_start_time = start_time
        elif event_name == "path":
            media_data["is_stream"] = False
            # if its ytdl://scheme
            if event_data.startswith("ytdl://"):
                media_data[event_name] = event_data.lstrip("ytdl://")
                media_data["is_stream"] = True
                continue
            if _is_urlish(event_data):
                media_data[event_name] = event_data
                media_data["is_stream"] = True
                continue
            # test if this is an absolute path
            if event_data.startswith("/"):
                media_data[event_name] = event_data
            else:
                # I think this is fine to do?
                full_path: str = os.path.join(working_dir, event_data)
                media_data[event_name] = full_path
        elif event_name == "metadata":
            # TODO: how to parse this better?
            media_data[event_name] = event_data
        elif event_name == "media-title":
            media_data["media_title"] = event_data
        elif event_name == "duration":
            # note: path is already set (if streaming, we may not get any duration)
            assert event_data is not None
            media_data[event_name] = float(event_data)
        elif event_name in ["seek", "paused", "resumed"]:
            if event_name == "paused":
                # if a pause event was received while mpv was still playing,
                # save when it was paused, we can calculate complete pause time
                # while this piece of media was playing by combining sequences of
                # pause times
                if is_playing:
                    is_playing = False
                    pause_start_time = dt_float
                else:
                    logger.debug("received pause event while paused?")
            elif event_name == "resumed":
                # if its currently paused, and we received a resume event
                if not is_playing:
                    is_playing = True
                    # if we know when it was paused, add how long it was paused to pause_duration
                    if pause_start_time is not None:
                        pause_duration = pause_duration + (dt_float - pause_start_time)
                        pause_start_time = None
                else:
                    # logger.warning("received resumed event while already playing?")
                    pass
            if event_data is not None and "percent-pos" in event_data:
                # if this is the 'resumed' event that happens when a socket first
                # connects, dont add it to the actions
                # also makes sure we already don't have any events. That way, if I was
                # spamming pause/play for some reason, when this was first launched or paused
                # some song immediately, those events are still used as normal
                #
                # its just the initial 'resumed' event that happens that is ignored
                if not (
                    event_name == "resumed"
                    and start_time is not None
                    and dt_float - start_time <= SCAN_TIME
                    and len(actions) == 0
                ):
                    actions[dt_float] = (event_name, event_data["percent-pos"])
        elif event_name == "eof":
            # eof is *ALWAYS* before new data gets loaded in
            # if mpv is force quit, may not have an eof.
            # check after to make sure eof/mpv-quit/final-write
            # was the last item, else write out whatever
            # media_data has in the dict currently
            if not is_playing:
                pause_duration = pause_duration + (dt_float - pause_start_time)  # type: ignore[operator]
            media_data["end_time"] = dt_float
            media_data["pause_duration"] = pause_duration
            media_data["actions"] = actions
            pause_duration = 0
            yield media_data
            media_data = {}
            actions = {}
        elif event_name in ["mpv-quit", "final-write"]:
            # if this happened right after an eof, it can be ignored

            # if the eof didnt happen and mpv was quit manually, save
            # quit time as end_time
            if REQUIRED_KEYS.issubset(set(media_data)):
                # if I quit while it was paused
                if not is_playing and pause_start_time is not None:
                    pause_duration = pause_duration + (dt_float - pause_start_time)
                media_data["end_time"] = dt_float
                media_data["pause_duration"] = pause_duration
                media_data["actions"] = actions
                yield media_data
            return
        else:
            logger.warning(f"Unexpected event name {event_name}")

    if len(media_data) != 0:
        # if we have enough of the fields in the namedtuple, then this isnt
        # a corrupted file, its one that didnt have an eof/had events
        # after an eof for some reason
        if not REQUIRED_KEYS.issubset(set(media_data)):
            logger.debug("Ignoring leftover data... {}".format(media_data))
        else:
            # if we got through all the keys, and this has been playing for at least a minute,
            # even though this is sorta broken, log it anyways
            if most_recent_time - media_data["start_time"] > 60:
                # if it crashed while it was paused
                if not is_playing and pause_start_time is not None:
                    pause_duration = pause_duration + (
                        most_recent_time - pause_start_time
                    )
                logger.debug(
                    "slightly broken, but yielding anyways... {}".format(media_data)
                )
                media_data["end_time"] = most_recent_time
                media_data["pause_duration"] = pause_duration
                media_data["actions"] = actions
                yield media_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from setuptools import setup, find_packages

requirements = Path("requirements.txt").read_text().strip().splitlines()

long_description = Path("README.md").read_text()

pkg = "mpv_history_daemon"
setup(
    author="Sean Breckenridge",
    author_email="seanbrecke@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: End Users/Desktop",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
    description="Daemon which connects to active mpv instances, saving a history of what I watch/listen to",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    name=pkg,
    packages=find_packages(include=[pkg]),
    package_data={pkg: ["py.typed"]},
    entry_points={
        "console_scripts": ["mpv-history-daemon = mpv_history_daemon.__main__:cli"]
    },
    license="http://www.apache.org/licenses/LICENSE-2.0",
    scripts=["bin/mpv_history_daemon_restart"],
    url="https://github.com/seanbreckenridge/mpv-history-daemon",
    version="0.1.7",
)
